from collections import defaultdict
from datetime import datetime
from http import HTTPStatus
from typing import Mapping, List, Callable
from typing import Optional, Dict, Awaitable

from fastapi import APIRouter, FastAPI, Header
from pydantic import BaseModel
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.responses import Response

from pyctuator.environment.environment_provider import EnvironmentData
from pyctuator.health.health_provider import HealthSummary
from pyctuator.httptrace import TraceRecord, TraceRequest, TraceResponse
from pyctuator.httptrace.http_tracer import Traces
from pyctuator.impl import SBA_V2_CONTENT_TYPE
from pyctuator.impl.pyctuator_impl import PyctuatorImpl
from pyctuator.impl.pyctuator_router import PyctuatorRouter, EndpointsData
from pyctuator.logging.pyctuator_logging import LoggersData, LoggerLevels
from pyctuator.metrics.metrics_provider import Metric, MetricNames
from pyctuator.threads.thread_dump_provider import ThreadDump


class FastApiLoggerItem(BaseModel):
    configuredLevel: Optional[str]


# pylint: disable=too-many-locals
class FastApiPyctuator(PyctuatorRouter):

    # pylint: disable=unused-variable
    def __init__(
            self,
            app: FastAPI,
            pyctuator_impl: PyctuatorImpl,
            include_in_openapi_schema: bool = False,
            customizer: Callable[[APIRouter], None] = None
    ) -> None:
        super().__init__(app, pyctuator_impl)
        router = APIRouter()
        if customizer:
            customizer(router)

        @router.get("/", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_endpoints() -> EndpointsData:
            return self.get_endpoints_data()

        @router.options("/env", include_in_schema=include_in_openapi_schema)
        @router.options("/info", include_in_schema=include_in_openapi_schema)
        @router.options("/health", include_in_schema=include_in_openapi_schema)
        @router.options("/metrics", include_in_schema=include_in_openapi_schema)
        @router.options("/loggers", include_in_schema=include_in_openapi_schema)
        @router.options("/dump", include_in_schema=include_in_openapi_schema)
        @router.options("/threaddump", include_in_schema=include_in_openapi_schema)
        @router.options("/logfile", include_in_schema=include_in_openapi_schema)
        @router.options("/trace", include_in_schema=include_in_openapi_schema)
        @router.options("/httptrace", include_in_schema=include_in_openapi_schema)
        def options() -> None:
            """
            Spring boot admin, after registration, issues multiple OPTIONS request to the monitored application in order
            to determine the supported capabilities (endpoints).
            Here we "acknowledge" that env, info and health are supported.
            The "include_in_schema=False" is used to prevent from these OPTIONS endpoints to show up in the
            documentation.
            """

        @router.get("/env", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_environment() -> EnvironmentData:
            return pyctuator_impl.get_environment()

        @router.get("/info", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_info() -> Dict:
            return pyctuator_impl.get_app_info()

        @router.get("/health", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_health() -> HealthSummary:
            return pyctuator_impl.get_health()

        @router.get("/metrics", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_metric_names() -> MetricNames:
            return pyctuator_impl.get_metric_names()

        @router.get("/metrics/{metric_name}", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_metric_measurement(metric_name: str) -> Metric:
            return pyctuator_impl.get_metric_measurement(metric_name)

        # Retrieving All Loggers
        @router.get("/loggers", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_loggers() -> LoggersData:
            return pyctuator_impl.logging.get_loggers()

        @router.post("/loggers/{logger_name}", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def set_logger_level(item: FastApiLoggerItem, logger_name: str) -> Dict:
            pyctuator_impl.logging.set_logger_level(logger_name, item.configuredLevel)
            return {}

        @router.get("/loggers/{logger_name}", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_logger(logger_name: str) -> LoggerLevels:
            return pyctuator_impl.logging.get_logger(logger_name)

        @router.get("/dump", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        @router.get("/threaddump", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_thread_dump() -> ThreadDump:
            return pyctuator_impl.get_thread_dump()

        @router.get("/logfile", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_logfile(range_header: str = Header(default=None,
                                                   alias="range")) -> Response:  # pylint: disable=redefined-builtin
            if not range_header:
                return Response(content=pyctuator_impl.logfile.log_messages.get_range())

            str_res, start, end = pyctuator_impl.logfile.get_logfile(range_header)

            my_res = Response(
                status_code=HTTPStatus.PARTIAL_CONTENT.value,
                content=str_res,
                headers={
                    "Content-Type": "text/html; charset=UTF-8",
                    "Accept-Ranges": "bytes",
                    "Content-Range": f"bytes {start}-{end}/{end}",
                })

            return my_res

        @router.get("/trace", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        @router.get("/httptrace", include_in_schema=include_in_openapi_schema, tags=["pyctuator"])
        def get_httptrace() -> Traces:
            return pyctuator_impl.http_tracer.get_httptrace()

        @app.middleware("http")
        async def intercept_requests_and_responses(
                request: Request,
                call_next: Callable[[Request], Awaitable[Response]]
        ) -> Response:
            request_time = datetime.now()
            response: Response = await call_next(request)
            response_time = datetime.now()

            # Set the SBA-V2 content type for responses from Pyctuator
            if request.url.path.startswith(self.pyctuator_impl.pyctuator_endpoint_path_prefix):
                response.headers["Content-Type"] = SBA_V2_CONTENT_TYPE

            # Record the request and response
            new_record = self._create_record(request, response, request_time, response_time)
            self.pyctuator_impl.http_tracer.add_record(record=new_record)

            return response

        app.include_router(router, prefix=pyctuator_impl.pyctuator_endpoint_path_prefix)

    def _create_headers_dictionary(self, headers: Headers) -> Mapping[str, List[str]]:
        headers_dict: Mapping[str, List[str]] = defaultdict(list)
        for (key, value) in headers.items():
            headers_dict[key].append(value)
        return headers_dict

    def _create_record(
            self,
            request: Request,
            response: Response,
            request_time: datetime,
            response_time: datetime,
    ) -> TraceRecord:
        new_record: TraceRecord = TraceRecord(
            request_time,
            None,
            None,
            TraceRequest(request.method, str(request.url), self._create_headers_dictionary(request.headers)),
            TraceResponse(response.status_code, self._create_headers_dictionary(response.headers)),
            int((response_time.timestamp() - request_time.timestamp()) * 1000),
        )
        return new_record



from dataclasses import dataclass
from datetime import datetime
from typing import List, Mapping, Optional


@dataclass
class TraceResponse:
    status: int
    headers: Mapping[str, List[str]]


@dataclass
class TraceRequest:
    method: str
    uri: str
    headers: Mapping[str, List[str]]


@dataclass
class Session:
    id: str


@dataclass
class Principal:
    name: str


@dataclass
class TraceRecord:
    timestamp: datetime
    principal: Optional[Principal]
    session: Optional[Session]
    request: TraceRequest
    response: TraceResponse
    timeTaken: int


@dataclass
class Traces:
    traces: List[TraceRecord]



# pylint: disable=import-outside-toplevel
import atexit
import importlib.util
import logging
import ssl
from datetime import datetime, timezone
from typing import Any, Optional, Dict, Callable

# A note about imports: this module ensure that only relevant modules are imported.
# For example, if the webapp is a Flask webapp, we do not want to import FastAPI, and vice versa.
# To do that, all imports are in conditional branches after detecting which frameworks are installed.
# DO NOT add any web-framework-dependent imports to the global scope.
from pyctuator.auth import Auth
from pyctuator.environment.custom_environment_provider import CustomEnvironmentProvider
from pyctuator.environment.os_env_variables_impl import OsEnvironmentVariableProvider
from pyctuator.health.diskspace_health_impl import DiskSpaceHealthProvider
from pyctuator.health.health_provider import HealthProvider
from pyctuator.metrics.memory_metrics_impl import MemoryMetricsProvider
from pyctuator.metrics.thread_metrics_impl import ThreadMetricsProvider
from pyctuator.impl.pyctuator_impl import PyctuatorImpl, AppInfo, BuildInfo, GitInfo, GitCommitInfo, AppDetails
from pyctuator.impl.spring_boot_admin_registration import BootAdminRegistrationHandler

default_logfile_format = '%(asctime)s  %(levelname)-5s %(process)d -- [%(threadName)s] %(module)s: %(message)s'


class Pyctuator:
    # pylint: disable=too-many-locals
    def __init__(
            self,
            app: Any,
            app_name: str,
            app_url: str,
            pyctuator_endpoint_url: str,
            registration_url: Optional[str],
            registration_auth: Optional[Auth] = None,
            app_description: Optional[str] = None,
            registration_interval_sec: float = 10,
            free_disk_space_down_threshold_bytes: int = 1024 * 1024 * 100,
            logfile_max_size: int = 10000,
            logfile_formatter: str = default_logfile_format,
            auto_deregister: bool = True,
            metadata: Optional[dict] = None,
            additional_app_info: Optional[dict] = None,
            ssl_context: Optional[ssl.SSLContext] = None,
            customizer: Optional[Callable] = None,
    ) -> None:
        """The entry point for integrating pyctuator with a web-frameworks such as FastAPI and Flask.

        Given an application built on top of a supported web-framework, it'll add to the application the REST API
         endpoints that required for Spring Boot Admin to monitor and manage the application.

        Pyctuator currently supports application built on top of FastAPI and Flask. The type of first argument, app is
         specific to the target web-framework:

        * FastAPI - `app` is an instance of `fastapi.applications.FastAPI`

        * Flask - `app` is an instance of `flask.app.Flask`

        * aiohttp - `app` is an instance of `aiohttp.web.Application`

        * Tornado - `app` is an instance of `tornado.web.Application`

        :param app: an instance of a supported web-framework with which the pyctuator endpoints will be registered
        :param app_name: the application's name that will be presented in the "Info" section in boot-admin
        :param app_description: a description that will be presented in the "Info" section in boot-admin
        :param app_url: the full URL of the application being monitored which will be displayed in spring-boot-admin, we
         recommend this URL to be accessible by those who manage the application (i.e. don't use "http://localhost..."
         as it is only accessible from within the application's host)
        :param pyctuator_endpoint_url: the public URL from which Pyctuator REST API will be accessible, used for
         registering the application with spring-boot-admin, must be accessible from spring-boot-admin server (i.e.
         don't use http://localhost:8080/... unless spring-boot-admin is running on the same host as the monitored
         application)
        :param registration_url: the spring-boot-admin endpoint to which registration requests must be posted
        :param registration_auth: optional authentication details to use when registering with spring-boot-admin
        :param registration_interval_sec: how often pyctuator will renew its registration with spring-boot-admin
        :param free_disk_space_down_threshold_bytes: amount of free space in bytes in "./" (the application's current
         working directory) below which the built-in disk-space health-indicator will fail
        :param auto_deregister: if true, pyctuator will automatically deregister from SBA during shutdown, needed for
        example when running in k8s so every time a new pod is created it is assigned a different IP address, resulting
        with SBA showing "offline" instances
        :param metadata: optional metadata key-value pairs that are displayed in SBA main page of an instance
        :param additional_app_info: additional arbitrary information to add to the application's "Info" section
        :param ssl_context: optional SSL context to be used when registering with SBA
        :param customizer: a function that can customize the integration with the web-framework which is therefore web-
         framework specific. For FastAPI, the function receives pyctuator's APIRouter allowing to add "dependencies" and
         anything else that's provided by the router. See fastapi_with_authentication_example_app.py
        """

        self.auto_deregister = auto_deregister
        start_time = datetime.now(timezone.utc)

        # Instantiate an instance of PyctuatorImpl which abstracts the state and logic of the pyctuator
        self.pyctuator_impl = PyctuatorImpl(
            AppInfo(app=AppDetails(name=app_name, description=app_description)),
            pyctuator_endpoint_url,
            logfile_max_size,
            logfile_formatter,
            additional_app_info,
        )

        # Register default health/metrics/environment providers
        self.pyctuator_impl.register_environment_provider(OsEnvironmentVariableProvider())
        self.pyctuator_impl.register_health_providers(DiskSpaceHealthProvider(free_disk_space_down_threshold_bytes))
        self.pyctuator_impl.register_metrics_provider(MemoryMetricsProvider())
        self.pyctuator_impl.register_metrics_provider(ThreadMetricsProvider())

        self.boot_admin_registration_handler: Optional[BootAdminRegistrationHandler] = None

        self.metadata = metadata
        self.ssl_context = ssl_context

        root_logger = logging.getLogger()
        # If application did not initiate logging module, add default handler to root logger
        # logging.info implicitly calls logging.basicConfig(), see logging.basicConfig in Python's documentation.
        if not root_logger.hasHandlers():
            logging.info("Logging not configured, using logging.basicConfig()")

        root_logger.addHandler(self.pyctuator_impl.logfile.log_messages)

        # Find and initialize an integration layer between the web-framework adn pyctuator
        framework_integrations: Dict[str, Callable[[Any, PyctuatorImpl, Optional[Callable]], bool]] = {
            "flask": self._integrate_flask,
            "fastapi": self._integrate_fastapi,
            "aiohttp": self._integrate_aiohttp,
            "tornado": self._integrate_tornado
        }
        for framework_name, framework_integration_function in framework_integrations.items():
            if self._is_framework_installed(framework_name):
                logging.debug("Framework %s is installed, trying to integrate with it", framework_name)
                success = framework_integration_function(app, self.pyctuator_impl, customizer)
                if success:
                    logging.debug("Integrated with framework %s", framework_name)
                    if registration_url is not None:
                        self.boot_admin_registration_handler = BootAdminRegistrationHandler(
                            registration_url,
                            registration_auth,
                            app_name,
                            self.pyctuator_impl.pyctuator_endpoint_url,
                            start_time,
                            app_url,
                            registration_interval_sec,
                            self.metadata,
                            self.ssl_context,
                        )

                        # Deregister from SBA on exit
                        if self.auto_deregister:
                            atexit.register(self.boot_admin_registration_handler.deregister_from_admin_server)

                        self.boot_admin_registration_handler.start()
                    return

        # Fail in case no framework was found for the target app
        raise EnvironmentError("No framework was found that is matching the target app "
                               "(is it properly installed and imported?)")

    def stop(self) -> None:
        if self.boot_admin_registration_handler:
            self.boot_admin_registration_handler.stop()
        self.boot_admin_registration_handler = None

    def set_secret_scrubber(self, secret_scrubber: Callable[[Dict], Dict]) -> None:
        """Overrides the default secret scrubber with a custom one. See SecretScrubber for example scrubber."""
        self.pyctuator_impl.set_secret_scrubber(secret_scrubber)

    def register_environment_provider(self, name: str, env_provider: Callable[[], Dict]) -> None:
        self.pyctuator_impl.register_environment_provider(CustomEnvironmentProvider(name, env_provider))

    def register_health_provider(self, provider: HealthProvider) -> None:
        self.pyctuator_impl.register_health_providers(provider)

    def set_git_info(self, commit: str, time: datetime, branch: Optional[str] = None) -> None:
        self.pyctuator_impl.set_git_info(GitInfo(GitCommitInfo(time, commit), branch))

    def set_build_info(
            self,
            artifact: Optional[str] = None,
            group: Optional[str] = None,
            name: Optional[str] = None,
            version: Optional[str] = None,
            time: Optional[datetime] = None,
    ) -> None:
        self.pyctuator_impl.set_build_info(BuildInfo(name, artifact, group, version, time))

    def _is_framework_installed(self, framework_name: str) -> bool:
        return importlib.util.find_spec(framework_name) is not None

    def _integrate_fastapi(self, app: Any, pyctuator_impl: PyctuatorImpl, customizer: Optional[Callable]) -> bool:
        """
        This method should only be called if we detected that FastAPI is installed.
        It will then check whether the given app is a FastAPI app, and if so - it will add the Pyctuator
        endpoints to it.
        """
        from fastapi import FastAPI
        if isinstance(app, FastAPI):
            from pyctuator.impl.fastapi_pyctuator import FastApiPyctuator
            FastApiPyctuator(app, pyctuator_impl, False, customizer)
            return True
        return False

    # pylint: disable=unused-argument
    def _integrate_flask(self, app: Any, pyctuator_impl: PyctuatorImpl, customizer: Optional[Callable]) -> bool:
        """
        This method should only be called if we detected that Flask is installed.
        It will then check whether the given app is a Flask app, and if so - it will add the Pyctuator
        endpoints to it.
        """
        from flask import Flask
        if isinstance(app, Flask):
            from pyctuator.impl.flask_pyctuator import FlaskPyctuator
            FlaskPyctuator(app, pyctuator_impl)
            return True
        return False

    # pylint: disable=unused-argument
    def _integrate_aiohttp(self, app: Any, pyctuator_impl: PyctuatorImpl, customizer: Optional[Callable]) -> bool:
        """
        This method should only be called if we detected that aiohttp is installed.
        It will then check whether the given app is a aiohttp app, and if so - it will add the Pyctuator
        endpoints to it.
        """
        from aiohttp.web import Application
        if isinstance(app, Application):
            from pyctuator.impl.aiohttp_pyctuator import AioHttpPyctuator
            AioHttpPyctuator(app, pyctuator_impl)
            return True
        return False

    # pylint: disable=unused-argument
    def _integrate_tornado(self, app: Any, pyctuator_impl: PyctuatorImpl, customizer: Optional[Callable]) -> bool:
        """
        This method should only be called if we detected that tornado is installed.
        It will then check whether the given app is a tornado app, and if so - it will add the Pyctuator
        endpoints to it.
        """
        from tornado.web import Application
        if isinstance(app, Application):
            from pyctuator.impl.tornado_pyctuator import TornadoHttpPyctuator
            TornadoHttpPyctuator(app, pyctuator_impl)
            return True
        return False



import http.client
import json
import logging
import os
import ssl
import threading
import urllib.parse
from base64 import b64encode
from datetime import datetime
from http.client import HTTPConnection, HTTPResponse
from typing import Optional, Dict

from pyctuator.auth import Auth, BasicAuth


# pylint: disable=too-many-instance-attributes
class BootAdminRegistrationHandler:

    def __init__(
            self,
            registration_url: str,
            registration_auth: Optional[Auth],
            application_name: str,
            pyctuator_base_url: str,
            start_time: datetime,
            service_url: str,
            registration_interval_sec: float,
            application_metadata: Optional[dict] = None,
            ssl_context: Optional[ssl.SSLContext] = None,
    ) -> None:
        self.registration_url = registration_url
        self.registration_auth = registration_auth
        self.application_name = application_name
        self.pyctuator_base_url = pyctuator_base_url
        self.start_time = start_time
        self.service_url = service_url if service_url.endswith("/") else service_url + "/"
        self.registration_interval_sec = registration_interval_sec
        self.instance_id = None
        self.application_metadata = application_metadata if application_metadata else {}
        self.ssl_context = ssl_context

        self.should_continue_registration_schedule: bool = False
        self.disable_certificate_validation_for_https_registration: bool = \
            os.getenv("PYCTUATOR_REGISTRATION_NO_CERT") is not None

    def _schedule_next_registration(self, registration_interval_sec: float) -> None:
        timer = threading.Timer(
            registration_interval_sec,
            self._register_with_admin_server,
            []
        )
        timer.setDaemon(True)
        timer.start()

    def _register_with_admin_server(self) -> None:
        # When waking up, make sure registration is still needed
        if not self.should_continue_registration_schedule:
            return

        registration_data = {
            "name": self.application_name,
            "managementUrl": self.pyctuator_base_url,
            "healthUrl": f"{self.pyctuator_base_url}/health",
            "serviceUrl": self.service_url,
            "metadata": {
                "startup": self.start_time.isoformat(),
                **self.application_metadata
            }
        }

        logging.debug("Trying to post registration data to %s: %s", self.registration_url, registration_data)

        conn: Optional[HTTPConnection] = None
        try:
            headers = {"Content-type": "application/json"}
            self.authenticate(headers)

            response = self._http_request(self.registration_url, "POST", headers, json.dumps(registration_data))

            if response.status < 200 or response.status >= 300:
                logging.warning("Failed registering with boot-admin, got %s - %s", response.status, response.read())
            else:
                self.instance_id = json.loads(response.read().decode('utf-8'))["id"]

        except Exception as e:  # pylint: disable=broad-except
            logging.warning("Failed registering with boot-admin, %s (%s)", e, type(e))

        finally:
            if conn:
                conn.close()

        # Schedule the next registration unless asked to abort
        if self.should_continue_registration_schedule:
            self._schedule_next_registration(self.registration_interval_sec)

    def deregister_from_admin_server(self) -> None:
        if self.instance_id is None:
            return

        headers = {}
        self.authenticate(headers)

        deregistration_url = f"{self.registration_url}/{self.instance_id}"
        logging.info("Deregistering from %s", deregistration_url)

        conn: Optional[HTTPConnection] = None
        try:
            response = self._http_request(deregistration_url, "DELETE", headers)

            if response.status < 200 or response.status >= 300:
                logging.warning("Failed deregistering from boot-admin, got %s - %s", response.status, response.read())

        except Exception as e:  # pylint: disable=broad-except
            logging.warning("Failed deregistering from boot-admin, %s (%s)", e, type(e))

        finally:
            if conn:
                conn.close()

    def authenticate(self, headers: Dict) -> None:
        if isinstance(self.registration_auth, BasicAuth):
            password = self.registration_auth.password if self.registration_auth.password else ""
            authorization_string = self.registration_auth.username + ":" + password
            encoded_authorization: str = b64encode(bytes(authorization_string, "utf-8")).decode("ascii")
            headers["Authorization"] = f"Basic {encoded_authorization}"

    def start(self, initial_delay_sec: float = None) -> None:
        logging.info("Starting recurring registration of %s with %s",
                     self.pyctuator_base_url, self.registration_url)
        self.should_continue_registration_schedule = True
        self._schedule_next_registration(initial_delay_sec or self.registration_interval_sec)

    def stop(self) -> None:
        logging.info("Stopping recurring registration")
        self.should_continue_registration_schedule = False

    def _http_request(self, url: str, method: str, headers: Dict[str, str], body: Optional[str] = None) -> HTTPResponse:
        url_parts = urllib.parse.urlsplit(url)
        if url_parts.scheme == "http":
            conn = http.client.HTTPConnection(url_parts.hostname, url_parts.port)
        elif url_parts.scheme == "https":
            context = self.ssl_context
            if not context and self.disable_certificate_validation_for_https_registration:
                context = ssl.SSLContext()
                context.verify_mode = ssl.CERT_NONE
            conn = http.client.HTTPSConnection(url_parts.hostname, url_parts.port, context=context)
        else:
            raise ValueError(f"Unknown scheme in {url}")

        conn.request(
            method,
            url_parts.path,
            body=body,
            headers=headers,
        )
        return conn.getresponse()



import importlib.util
from dataclasses import dataclass
from typing import Optional

from sqlalchemy.engine import Engine

from pyctuator.health.health_provider import HealthProvider, HealthStatus, Status, HealthDetails


@dataclass
class DbHealthDetails(HealthDetails):
    engine: str
    failure: Optional[str] = None


@dataclass
class DbHealthStatus(HealthStatus):
    status: Status
    details: DbHealthDetails


class DbHealthProvider(HealthProvider):

    def __init__(self, engine: Engine) -> None:
        super().__init__()
        self.engine = engine

    def is_supported(self) -> bool:
        return importlib.util.find_spec("sqlalchemy") is not None

    def get_name(self) -> str:
        return "db"

    def get_health(self) -> DbHealthStatus:
        try:
            conn = self.engine.raw_connection()
            if self.engine.dialect.do_ping(conn):
                return DbHealthStatus(status=Status.UP, details=DbHealthDetails(self.engine.name))

            return DbHealthStatus(
                status=Status.UNKNOWN,
                details=DbHealthDetails(self.engine.name, f"Pinging failed"))

        except Exception as e:  # pylint: disable=broad-except
            return DbHealthStatus(status=Status.DOWN, details=DbHealthDetails(self.engine.name, str(e)))



# pylint: disable=import-outside-toplevel
import importlib.util
from dataclasses import dataclass

from pyctuator.health.health_provider import HealthProvider, HealthDetails, HealthStatus, Status


@dataclass
class DiskSpaceHealthDetails(HealthDetails):
    total: int
    free: int
    threshold: int


@dataclass
class DiskSpaceHealth(HealthStatus):
    status: Status
    details: DiskSpaceHealthDetails


class DiskSpaceHealthProvider(HealthProvider):

    def __init__(self, free_bytes_down_threshold: int) -> None:
        self.free_bytes_down_threshold = free_bytes_down_threshold

        if importlib.util.find_spec("psutil"):
            # psutil is optional and must only be imported if it is installed
            import psutil
            self.psutil = psutil
        else:
            self.psutil = None

    def is_supported(self) -> bool:
        return self.psutil is not None

    def get_name(self) -> str:
        return "diskSpace"

    def get_health(self) -> DiskSpaceHealth:
        usage = self.psutil.disk_usage(".")
        return DiskSpaceHealth(
            Status.UP if usage.free > self.free_bytes_down_threshold else Status.DOWN,
            DiskSpaceHealthDetails(usage.total, usage.free, self.free_bytes_down_threshold)
        )



import logging
from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass
class LoggerLevels:
    configuredLevel: str
    effectiveLevel: str


@dataclass
class LoggersData:
    levels: List[str]
    loggers: Dict[str, LoggerLevels]
    groups: Dict[str, LoggerLevels]


@dataclass
class LogLevelMapping:
    boot_admin_log_level: str
    python_log_level: int
    python_from_log_level: int


_log_level_mapping: List[LogLevelMapping] = [
    LogLevelMapping("DEBUG", logging.DEBUG, logging.NOTSET),
    LogLevelMapping("INFO", logging.INFO, logging.DEBUG),
    LogLevelMapping("WARN", logging.WARNING, logging.INFO),
    LogLevelMapping("ERROR", logging.ERROR, logging.WARNING),
    LogLevelMapping("OFF", logging.NOTSET, -1),
]


def _python_to_admin_log_level(log_level: int) -> str:
    for mapping in _log_level_mapping:
        if mapping.python_from_log_level < log_level <= mapping.python_log_level:
            return mapping.boot_admin_log_level

    # If log_level is unknown, simply return its string representation
    return str(log_level)


def _admin_to_python_log_level(log_level: str) -> int:
    log_level_mapping = next(mapping for mapping in _log_level_mapping if mapping.boot_admin_log_level == log_level)
    return log_level_mapping.python_log_level


class PyctuatorLogging:
    def set_logger_level(self, logger_name: str, logger_level: Optional[str]) -> None:
        logger = logging.getLogger(logger_name)
        level = logger_level or "OFF"

        if level == "OFF":
            logging.disable(logging.CRITICAL)  # disable all logging calls of (CRITICAL) severity lvl and below
            logger.setLevel(0)
        else:
            logger.setLevel(_admin_to_python_log_level(level))
        logging.debug("Setting logger '%s' level to %s", logger_name, level)

    def get_loggers(self) -> LoggersData:
        level_names = [mapping.boot_admin_log_level for mapping
                       in _log_level_mapping
                       if mapping.boot_admin_log_level != "OFF"]

        loggers = {}
        for logger_dict_member in logging.root.manager.loggerDict:  # type: ignore
            logger_inst = logging.getLogger(logger_dict_member)
            level = _python_to_admin_log_level(logger_inst.level)
            loggers[logger_inst.name] = LoggerLevels(level, level)

        return LoggersData(levels=level_names, loggers=loggers, groups={})

    def get_logger(self, logger_name: str) -> LoggerLevels:
        logger = logging.getLogger(logger_name)
        level = _python_to_admin_log_level(logger.level)
        return LoggerLevels(level, level)



# pylint: disable=import-outside-toplevel
import importlib.util
from typing import List

from pyctuator.metrics.metrics_provider import MetricsProvider, Metric, Measurement

PREFIX = "thread."
THREAD_COUNT = PREFIX + "count"


class ThreadMetricsProvider(MetricsProvider):
    def __init__(self) -> None:
        if importlib.util.find_spec("psutil"):
            # psutil is optional and must only be imported if it is installed
            import psutil
            self.process = psutil.Process()
        else:
            self.process = None

    def get_prefix(self) -> str:
        return PREFIX

    def get_supported_metric_names(self) -> List[str]:
        return [THREAD_COUNT] if self.process else []

    def get_metric(self, metric_name: str) -> Metric:
        measurements = [Measurement("COUNT", self.process.num_threads())] if self.process else []
        return Metric(metric_name, None, "Integer", measurements, [])



import dataclasses
import json
from datetime import datetime, timedelta
from functools import partial
from http import HTTPStatus
from typing import Any, Optional, Callable, Mapping, List

from tornado.httputil import HTTPHeaders
from tornado.web import Application, RequestHandler

from pyctuator.httptrace import TraceRecord, TraceRequest, TraceResponse
from pyctuator.impl import SBA_V2_CONTENT_TYPE
from pyctuator.impl.pyctuator_impl import PyctuatorImpl
from pyctuator.impl.pyctuator_router import PyctuatorRouter


# pylint: disable=abstract-method
class AbstractPyctuatorHandler(RequestHandler):
    pyctuator_router: Optional[PyctuatorRouter] = None
    dumps: Optional[Callable[[Any], str]] = None

    def initialize(self) -> None:
        self.pyctuator_router = self.application.settings.get("pyctuator_router")
        self.dumps = self.application.settings.get("custom_dumps")
        self.set_header("Content-Type", SBA_V2_CONTENT_TYPE)

    def options(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write("")


class PyctuatorHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.get_endpoints_data()))


# GET /env
class EnvHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.get_environment()))


# GET /info
class InfoHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.get_app_info()))


# GET /health
class HealthHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.get_health()))


# GET /metrics
class MetricsHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.get_metric_names()))


# GET "/metrics/{metric_name}"
class MetricsNameHandler(AbstractPyctuatorHandler):
    def get(self, metric_name: str) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.get_metric_measurement(metric_name)))


# GET /loggers
class LoggersHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.logging.get_loggers()))


# GET /loggers/{logger_name}
# POST /loggers/{logger_name}
class LoggersNameHandler(AbstractPyctuatorHandler):
    def get(self, logger_name: str) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.logging.get_logger(logger_name)))

    def post(self, logger_name: str) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        body_str = self.request.body.decode("utf-8")
        body = json.loads(body_str)
        self.pyctuator_router.pyctuator_impl.logging.set_logger_level(logger_name, body.get("configuredLevel", None))
        self.write("")


# GET /threaddump
class ThreadDumpHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.get_thread_dump()))


# GET /logfile
class LogFileHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None

        range_header = self.request.headers.get("range")
        if not range_header:
            self.write(f"{self.pyctuator_router.pyctuator_impl.logfile.log_messages.get_range()}")

        else:
            str_res, start, end = self.pyctuator_router.pyctuator_impl.logfile.get_logfile(range_header)
            self.set_status(HTTPStatus.PARTIAL_CONTENT.value)
            self.add_header("Content-Type", "text/html; charset=UTF-8")
            self.add_header("Accept-Ranges", "bytes")
            self.add_header("Content-Range", f"bytes {start}-{end}/{end}")
            self.write(str_res)


# GET /httptrace
class HttpTraceHandler(AbstractPyctuatorHandler):
    def get(self) -> None:
        assert self.pyctuator_router is not None
        assert self.dumps is not None
        self.write(self.dumps(self.pyctuator_router.pyctuator_impl.http_tracer.get_httptrace()))


# pylint: disable=too-many-locals,unused-argument
class TornadoHttpPyctuator(PyctuatorRouter):
    def __init__(self, app: Application, pyctuator_impl: PyctuatorImpl) -> None:
        super().__init__(app, pyctuator_impl)

        custom_dumps = partial(
            json.dumps, default=self._custom_json_serializer
        )

        app.settings.setdefault("pyctuator_router", self)
        app.settings.setdefault("custom_dumps", custom_dumps)

        # Register a log-function that records request and response in traces and than delegates to the original func
        self.delegate_log_function = app.settings.get("log_function")
        app.settings.setdefault("log_function", self._intercept_request_and_response)

        app.add_handlers(
            ".*$",
            [
                (r"/pyctuator", PyctuatorHandler),
                (r"/pyctuator/env", EnvHandler),
                (r"/pyctuator/info", InfoHandler),
                (r"/pyctuator/health", HealthHandler),
                (r"/pyctuator/metrics", MetricsHandler),
                (r"/pyctuator/metrics/(?P<metric_name>.*$)", MetricsNameHandler),
                (r"/pyctuator/loggers", LoggersHandler),
                (r"/pyctuator/loggers/(?P<logger_name>.*$)", LoggersNameHandler),
                (r"/pyctuator/dump", ThreadDumpHandler),
                (r"/pyctuator/threaddump", ThreadDumpHandler),
                (r"/pyctuator/logfile", LogFileHandler),
                (r"/pyctuator/trace", HttpTraceHandler),
                (r"/pyctuator/httptrace", HttpTraceHandler),
            ]
        )

    def _intercept_request_and_response(self, handler: RequestHandler) -> None:
        # Record the request and response
        record = TraceRecord(
            timestamp=datetime.now() - timedelta(seconds=handler.request.request_time()),
            principal=None,
            session=None,
            request=TraceRequest(
                method=handler.request.method or "",
                uri=handler.request.full_url(),
                headers=get_headers(handler.request.headers)
            ),
            response=TraceResponse(
                status=handler.get_status(),
                headers=get_headers(handler._headers)  # pylint: disable=protected-access
            ),
            timeTaken=int(handler.request.request_time() * 1000),
        )
        self.pyctuator_impl.http_tracer.add_record(record)

        if self.delegate_log_function:
            self.delegate_log_function(handler)

    def _custom_json_serializer(self, value: Any) -> Any:
        if dataclasses.is_dataclass(value):
            return dataclasses.asdict(value)

        if isinstance(value, datetime):
            return str(value)
        return None


def get_headers(headers: HTTPHeaders) -> Mapping[str, List[str]]:
    """ Tornado's HTTPHeaders contains multiple entries of the same header name if multiple values were used, this
    function groups headers by header name. See documentation of `tornado.httputil.HTTPHeaders` """
    return {header.lower(): headers.get_list(header) for header in headers.keys()}



import logging
import re
from typing import Optional, Tuple

logfile_request_range_pattern = re.compile("bytes=(\\d*)-(\\d*)")


class LogMessageBuffer(logging.Handler):
    def __init__(self, max_size: int, formatter: str) -> None:
        super().__init__()
        self.setFormatter(logging.Formatter(formatter))
        self._max_size = max_size
        self._buffer: str = ""
        self._offset: int = 0

    def emit(self, record: logging.LogRecord) -> None:
        msg = self.format(record) + "\n"
        msg_len = len(msg)
        if len(self._buffer) + msg_len > self._max_size:
            self._buffer = self._buffer[-(self._max_size - msg_len):]
            self._offset += msg_len
        self._buffer += msg

    def get_range(self, start: Optional[int] = None, end: Optional[int] = None) -> str:
        start = start - self._offset if start else 0
        return self._buffer[start:end or len(self._buffer) - 1]

    def get_offset(self) -> int:
        return self._offset

    def get_offset_tuple(self, start: Optional[int], end: Optional[int]) -> Tuple[int, int]:
        res_start = self._offset + (start or 0)
        res_end = self._offset + (end or len(self._buffer))
        return res_start, res_end


class PyctuatorLogfile:
    def __init__(self, max_size: int, formatter: str) -> None:
        self.log_messages = LogMessageBuffer(max_size=max_size, formatter=formatter)

    def get_logfile(self, range_substring: str) -> Tuple[str, int, int]:
        logging.debug("Received logfile request with range header: %s", range_substring)

        start_str, end_str = logfile_request_range_pattern.match(range_substring).groups()
        start = int(start_str) if start_str.strip() else None
        end = int(end_str) if end_str.strip() else None

        str_res = self.log_messages.get_range(start, end)
        end = len(str_res) if (start is None) and end else end  # Handle 0-307200 initial range edge-case

        res_start, res_end = self.log_messages.get_offset_tuple(start, end)

        logging.debug(f"Returning logfile response with range header: bytes=%d-%d/%d", res_start, res_end, res_end)

        return str_res, res_start, res_end

    def get_log_buffer_offset(self) -> int:
        return self.log_messages.get_offset()



import os
from typing import Dict, Callable

from pyctuator.environment.environment_provider import PropertiesSource, PropertyValue, EnvironmentProvider


class OsEnvironmentVariableProvider(EnvironmentProvider):

    def get_properties_source(self, secret_scrubber: Callable[[Dict], Dict]) -> PropertiesSource:
        scrubbed_env = secret_scrubber(os.environ)  # type: ignore
        properties_dict = {key: PropertyValue(value) for (key, value) in scrubbed_env.items()}
        return PropertiesSource("systemEnvironment", properties_dict)



import importlib.util
from dataclasses import dataclass
from typing import Optional

from redis import Redis

from pyctuator.health.health_provider import HealthProvider, HealthStatus, Status, HealthDetails


@dataclass
class RedisHealthDetails(HealthDetails):
    version: Optional[str] = None
    mode: Optional[str] = None
    failure: Optional[str] = None


@dataclass
class RedisHealthStatus(HealthStatus):
    status: Status
    details: RedisHealthDetails


class RedisHealthProvider(HealthProvider):

    def __init__(self, redis: Redis) -> None:
        super().__init__()
        self.redis = redis

    def is_supported(self) -> bool:
        return importlib.util.find_spec("redis") is not None

    def get_name(self) -> str:
        return "redis"

    def get_health(self) -> RedisHealthStatus:
        try:
            info = self.redis.info()

            return RedisHealthStatus(
                status=Status.UP,
                details=RedisHealthDetails(
                    version=info["redis_version"],
                    mode=info["redis_mode"],
                ))
        except Exception as e:  # pylint: disable=broad-except
            return RedisHealthStatus(
                status=Status.DOWN,
                details=RedisHealthDetails(
                    failure=str(e)
                ))



from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import Mapping, Optional, List, Any, Callable, Dict


@dataclass
class PropertyValue:
    value: Any
    origin: Optional[str] = None


@dataclass
class PropertiesSource:
    name: str
    properties: Mapping[str, PropertyValue]


@dataclass
class EnvironmentData:
    activeProfiles: List[str]
    propertySources: List[PropertiesSource]


class EnvironmentProvider(ABC):

    @abstractmethod
    def get_properties_source(self, secret_scrubber: Callable[[Dict], Dict]) -> PropertiesSource:
        pass



import re
from typing import Any, Dict, Pattern

default_keys_to_scrub = re.compile("^(.*[^A-Za-z])?key([^A-Za-z].*)?$|.*secret.*|.*password.*|.*token.*", re.IGNORECASE)


class SecretScrubber:

    def __init__(self, keys_to_scrub: Pattern[str] = default_keys_to_scrub) -> None:
        self.keys_to_scrub = keys_to_scrub
        self.url_keys_to_scrub = re.compile(".*url.*", re.IGNORECASE)

    def scrub_secrets(self, mapping: Dict) -> Dict:
        """Scrubs secrets from a dictionary replacing them with stars

        :param mapping: a mapping with "primitive" values that may include secrets
        :return: a copy of the input mapping having all secrets replaced with stars
        """

        def scrub(key: Any, value: Any) -> Any:
            if self.keys_to_scrub.match(key):
                return "******"

            if self.url_keys_to_scrub.match(key):
                return re.sub(r"(.*//[^:]*:).*(@.*)", r"\1******\2", str(value))

            return value

        return {k: scrub(k, v) for (k, v) in mapping.items()}



from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import List, Optional


@dataclass
class MetricNames:
    names: List[str]


@dataclass
class Measurement:
    statistic: str  # one of TOTAL, TOTAL_TIME, COUNT, MAX, VALUE, UNKNOWN, ACTIVE_TASKS, DURATION
    value: float  # can be an int as well


@dataclass
class MetricTag:
    tag: str
    values: List[str]


@dataclass
class Metric:
    name: str
    description: Optional[str]
    baseUnit: str
    measurements: List[Measurement]
    availableTags: List[MetricTag]


class MetricsProvider(ABC):

    @abstractmethod
    def get_prefix(self) -> str:
        pass

    @abstractmethod
    def get_supported_metric_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_metric(self, metric_name: str) -> Metric:
        pass



import sys
import threading
from threading import Thread
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Any, Optional


@dataclass
class StackFrame:
    methodName: str
    fileName: str
    lineNumber: int
    className: Optional[str]
    nativeMethod: bool


@dataclass
class ThreadInfo:
    threadName: str
    threadId: Optional[int]
    daemon: bool
    suspended: bool
    threadState: str
    stackTrace: List[StackFrame]


@dataclass
class ThreadDump:
    threads: List[ThreadInfo]


class ThreadDumpProvider:

    # pylint: disable=protected-access
    def get_thread_dump(self) -> ThreadDump:
        frames: Dict[Any, Any] = sys._current_frames()
        return ThreadDump([
            self._extract_thread_info(frames, thread)
            for thread in threading.enumerate()
        ])

    def _extract_thread_info(self, frames: Dict[Any, Any], thread: Thread) -> ThreadInfo:
        return ThreadInfo(
            threadName=thread.name,
            threadId=thread.ident,
            daemon=thread.daemon,
            suspended=not thread.is_alive(),
            threadState=self._calc_thread_state(thread),
            stackTrace=self._build_thread_stack_trace(thread, frames),
        )

    def _build_thread_stack_trace(self, thread: Thread, frames: Dict[Any, Any]) -> List[StackFrame]:

        def guess_class_name() -> Optional[str]:
            """
            Tries to find a class name if one exists.
            Fails if the frame is not in a class, or if the method does not call itself "self"
            Does not support static and class methods.
            """
            try:
                return str(frame.f_locals["self"].__class__.__name__)
            except KeyError:
                return None

        stack_frames = []
        frame = frames[thread.ident] if thread.ident in frames else None
        while frame is not None:
            stack_frames.append(StackFrame(
                methodName=frame.f_code.co_name,
                fileName=Path(frame.f_code.co_filename).name,
                lineNumber=frame.f_lineno,
                className=guess_class_name(),
                nativeMethod=False
            ))
            frame = frame.f_back  # move one frame back
        return stack_frames

    def _calc_thread_state(self, thread: threading.Thread) -> str:
        if thread.ident and thread.ident < 0:
            return "NEW"
        return "RUNNABLE"



import re

_keys_to_scrub = re.compile(
    "^(.*[^A-Za-z])?key([^A-Za-z].*)?$|"
    ".*secret.*|"
    ".*password.*|"
    ".*token.*|"
    ".*authorization.*|"
    ".*authentication.*|"
    ".*cookie.*",
    re.IGNORECASE
)


def scrub_header_value(key: str, value: str) -> str:
    if _keys_to_scrub.match(key):
        return "******"

    return value



import collections
from typing import List, Mapping
from pyctuator.httptrace.http_header_scrubber import scrub_header_value

from pyctuator.httptrace import Traces, TraceRecord


class HttpTracer:
    def __init__(self) -> None:
        self.traces_list: collections.deque = collections.deque(maxlen=100)

    def get_httptrace(self) -> Traces:
        return Traces(list(self.traces_list))

    def add_record(self, record: TraceRecord) -> None:

        record.request.headers = self._scrub_and_normalize_headers(
            record.request.headers)
        record.response.headers = self._scrub_and_normalize_headers(
            record.response.headers)

        self.traces_list.append(record)

    def _scrub_and_normalize_headers(self, headers: Mapping[str, List[str]]) -> Mapping[str, List[str]]:
        return {header: [scrub_header_value(header, value) for value in values] for (header, values) in headers.items()}



import json
from collections import defaultdict
from datetime import datetime, date
from http import HTTPStatus
from typing import Dict, Tuple, Any, Mapping, List

from flask import Flask, Blueprint, request, jsonify, after_this_request
from flask import Response, make_response
from flask.json import JSONEncoder
from werkzeug.datastructures import Headers

from pyctuator.httptrace import TraceRecord, TraceRequest, TraceResponse
from pyctuator.impl import SBA_V2_CONTENT_TYPE
from pyctuator.impl.pyctuator_impl import PyctuatorImpl
from pyctuator.impl.pyctuator_router import PyctuatorRouter


class CustomJSONEncoder(JSONEncoder):
    """ Override Flask's JSON encoding of datetime to assure ISO format is used.

    By default, when Flask is rendering a response to JSON, it is formatting datetime, date and time according to
    RFC-822 which is different from the ISO format used by SBA.

    This encoder overrides the default datetime encoding and is only used by the Pyctuator blueprint so it shouldn't
    interfere with whatever encoding users are using.

    See https://stackoverflow.com/questions/43663552/keep-a-datetime-date-in-yyyy-mm-dd-format-when-using-flasks-jsonify
    """

    # pylint: disable=method-hidden
    def default(self, o: Any) -> Any:
        try:
            if isinstance(o, (datetime, date)):
                return o.isoformat()
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        return JSONEncoder.default(self, o)


class FlaskPyctuator(PyctuatorRouter):

    # pylint: disable=too-many-locals, unused-variable
    def __init__(
            self,
            app: Flask,
            pyctuator_impl: PyctuatorImpl

    ) -> None:
        super().__init__(app, pyctuator_impl)

        path_prefix: str = pyctuator_impl.pyctuator_endpoint_path_prefix
        flask_blueprint: Blueprint = Blueprint("flask_blueprint", "pyctuator", )
        flask_blueprint.json_encoder = CustomJSONEncoder

        @app.before_request
        def intercept_requests_and_responses() -> None:
            request_time = datetime.now()

            @after_this_request
            def after_response(response: Response) -> Response:
                response_time = datetime.now()

                # Set the SBA-V2 content type for responses from Pyctuator
                if request.path.startswith(self.pyctuator_impl.pyctuator_endpoint_path_prefix):
                    response.headers["Content-Type"] = SBA_V2_CONTENT_TYPE

                # Record the request and response
                self.record_request_and_response(response, request_time, response_time)
                return response

        @flask_blueprint.route("/")
        def get_endpoints() -> Any:
            return jsonify(self.get_endpoints_data())

        @flask_blueprint.route("/env")
        def get_environment() -> Any:
            return jsonify(pyctuator_impl.get_environment())

        @flask_blueprint.route("/info")
        def get_info() -> Any:
            return jsonify(pyctuator_impl.get_app_info())

        @flask_blueprint.route("/health")
        def get_health() -> Any:
            return jsonify(pyctuator_impl.get_health())

        @flask_blueprint.route("/metrics")
        def get_metric_names() -> Any:
            return jsonify(pyctuator_impl.get_metric_names())

        @flask_blueprint.route("/metrics/<metric_name>")
        def get_metric_measurement(metric_name: str) -> Any:
            return jsonify(pyctuator_impl.get_metric_measurement(metric_name))

        # Retrieving All Loggers
        @flask_blueprint.route("/loggers")
        def get_loggers() -> Any:
            return jsonify(pyctuator_impl.logging.get_loggers())

        @flask_blueprint.route("/loggers/<logger_name>", methods=['POST'])
        def set_logger_level(logger_name: str) -> Dict:
            request_dict = json.loads(request.data)
            pyctuator_impl.logging.set_logger_level(logger_name, request_dict.get("configuredLevel", None))
            return {}

        @flask_blueprint.route("/loggers/<logger_name>")
        def get_logger(logger_name: str) -> Any:
            return jsonify(pyctuator_impl.logging.get_logger(logger_name))

        @flask_blueprint.route("/threaddump")
        @flask_blueprint.route("/dump")
        def get_thread_dump() -> Any:
            return jsonify(pyctuator_impl.get_thread_dump())

        @flask_blueprint.route("/logfile")
        def get_logfile() -> Tuple[Response, int]:
            range_header: str = request.headers.environ.get('HTTP_RANGE')
            if not range_header:
                response: Response = make_response(pyctuator_impl.logfile.log_messages.get_range())
                return response, HTTPStatus.OK

            str_res, start, end = pyctuator_impl.logfile.get_logfile(range_header)

            resp: Response = make_response(str_res)
            resp.headers["Content-Type"] = "text/html; charset=UTF-8"
            resp.headers["Accept-Ranges"] = "bytes"
            resp.headers["Content-Range"] = f"bytes {start}-{end}/{end}"

            return resp, HTTPStatus.PARTIAL_CONTENT

        @flask_blueprint.route("/trace")
        @flask_blueprint.route("/httptrace")
        def get_httptrace() -> Any:
            return jsonify(pyctuator_impl.http_tracer.get_httptrace())

        app.register_blueprint(flask_blueprint, url_prefix=path_prefix)

    def _create_headers_dictionary_flask(self, headers: Headers) -> Mapping[str, List[str]]:
        headers_dict: Mapping[str, List[str]] = defaultdict(list)
        for (key, value) in headers.items():
            headers_dict[key].append(value)
        return dict(headers_dict)

    def record_request_and_response(
            self,
            response: Response,
            request_time: datetime,
            response_time: datetime,
    ) -> None:
        new_record = TraceRecord(
            request_time,
            None,
            None,
            TraceRequest(request.method, str(request.url), self._create_headers_dictionary_flask(request.headers)),
            TraceResponse(response.status_code, self._create_headers_dictionary_flask(response.headers)),
            int((response_time.timestamp() - request_time.timestamp()) * 1000),
        )
        self.pyctuator_impl.http_tracer.add_record(record=new_record)



from abc import ABC
from dataclasses import dataclass
from enum import Enum

from typing import Mapping


class Status(str, Enum):
    UP = "UP"
    DOWN = "DOWN"
    UNKNOWN = "UNKNOWN"


@dataclass
class HealthDetails:
    pass


@dataclass
class HealthStatus:
    status: Status
    details: HealthDetails


@dataclass
class HealthSummary:
    status: Status
    details: Mapping[str, HealthStatus]


class HealthProvider(ABC):
    def is_supported(self) -> bool:
        pass

    def get_name(self) -> str:
        pass

    def get_health(self) -> HealthStatus:
        pass



# pylint: disable=import-outside-toplevel
import importlib.util
from typing import List

from pyctuator.metrics.metrics_provider import MetricsProvider, Metric, Measurement

PREFIX = "memory."


class MemoryMetricsProvider(MetricsProvider):
    def __init__(self) -> None:
        if importlib.util.find_spec("psutil"):
            # psutil is optional and must only be imported if it is installed
            import psutil
            self.process = psutil.Process()
        else:
            self.process = None

    def get_prefix(self) -> str:
        return PREFIX

    def get_supported_metric_names(self) -> List[str]:
        if not self.process:
            return []
        keys: List[str] = list(self.process.memory_info()._asdict().keys())
        return list(map(lambda metric: PREFIX + metric, list(keys)))

    def get_metric(self, metric_name: str) -> Metric:
        measurements: List[Measurement] = []
        if self.process:
            name = metric_name[len(PREFIX):]
            measurements = [Measurement("VALUE", getattr(self.process.memory_info(), name))]
        return Metric(metric_name, None, "bytes", measurements, [])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyctuator',
 'pyctuator.environment',
 'pyctuator.health',
 'pyctuator.httptrace',
 'pyctuator.impl',
 'pyctuator.logfile',
 'pyctuator.logging',
 'pyctuator.metrics',
 'pyctuator.threads']

package_data = \
{'': ['*']}

extras_require = \
{'aiohttp': ['aiohttp>=3.6.2,<4.0.0'],
 'db': ['sqlalchemy>=1.3,<2.0',
        'PyMySQL>=0.9.3,<0.10.0',
        'cryptography>=2.8,<4.0'],
 'fastapi': ['fastapi>=0.41.0,<0.42.0', 'uvicorn>=0.9.0,<0.10.0'],
 'flask': ['flask>=1.1,<2.0'],
 'psutil': ['psutil>=5.6,<6.0'],
 'redis': ['redis>=3.3,<4.0'],
 'tornado': ['tornado>=6.0.4,<7.0.0']}

setup_kwargs = {
    'name': 'pyctuator',
    'version': '0.18.1',
    'description': 'A Python implementation of the Spring Actuator API for popular web frameworks',
    'long_description': '[![PyPI](https://img.shields.io/pypi/v/pyctuator?color=green&style=plastic)](https://pypi.org/project/pyctuator/)\n[![build](https://github.com/SolarEdgeTech/pyctuator/workflows/build/badge.svg)](https://github.com/SolarEdgeTech/pyctuator/)\n[![Codecov](https://img.shields.io/codecov/c/github/SolarEdgeTech/pyctuator?style=plastic)](https://codecov.io/gh/SolarEdgeTech/pyctuator)\n\n# Pyctuator\n\nMonitor Python web apps using \n[Spring Boot Admin](https://github.com/codecentric/spring-boot-admin). \n\nPyctuator supports **[Flask](https://palletsprojects.com/p/flask/)**, **[FastAPI](https://fastapi.tiangolo.com/)**, **[aiohttp](docs.aiohttp.org)** and **[Tornado](https://www.tornadoweb.org/)**. **Django** support is planned as well.\n\nThe following video shows a FastAPI web app being monitored and controled using Spring Boot Admin.\n \n![Pyctuator Example](examples/images/Pyctuator_Screencast.gif)\n\nThe complete example can be found in [Advanced example](examples/Advanced/README.md).\n\n## Requirements\nPython 3.7+\n\nPyctuator has zero hard dependencies.\n\n## Installing\nInstall Pyctuator using pip: `pip3 install pyctuator`\n\n## Why?\nMany Java shops use Spring Boot as their main web framework for developing\nmicroservices. \nThese organizations often use Spring Actuator together with Spring Boot Admin\nto monitor their microservices\' status, gain access to applications\'\n state and configuration, manipulate log levels, etc.\n \nWhile Spring Boot is suitable for many use-cases, it is very common for organizations \nto also have a couple of Python microservices, as Python is often more suitable for \nsome types of applications. The most common examples are Data Science and Machine Learning\napplications.\n\nSetting up a proper monitoring tool for these microservices is a complex task, and might\nnot be justified for just a few Python microservices in a sea of Java microservices.\n\nThis is where Pyctuator comes in. It allows you to easily integrate your Python\nmicroservices into your existing Spring Boot Admin deployment.\n\n## Main Features\nPyctuator is a partial Python implementation of the \n[Spring Actuator API](https://docs.spring.io/spring-boot/docs/2.1.8.RELEASE/actuator-api/html/)  . \n\nIt currently supports the following Actuator features:\n\n* **Application details**\n* **Metrics**\n    * Memory usage\n    * Disk usage \n    * Custom metrics\n* **Health monitors**\n    * Built in MySQL health monitor\n    * Built in Redis health monitor\n    * Custom health monitors\n* **Environment**\n* **Loggers** - Easily change log levels during runtime\n* **Log file** - Tail the application\'s log file\n* **Thread dump** - See which threads are running\n* **HTTP traces** - Tail recent HTTP requests, including status codes and latency\n\n## Quickstart\nThe examples below show a minimal integration of **FastAPI**, **Flask** and **aiohttp** applications with **Pyctuator**.\n\nAfter installing Flask/FastAPI/aiohttp and Pyctuator, start by launching a local Spring Boot Admin instance:\n\n```sh\ndocker run --rm -p 8080:8080 --add-host=host.docker.internal:host-gateway michayaak/spring-boot-admin:2.2.3-1\n```\n\nThen go to `http://localhost:8080` to get to the web UI.\n\n### Flask\nThe following example is complete and should run as is.\n\n```python\nfrom flask import Flask\nfrom pyctuator.pyctuator import Pyctuator\n\napp_name = "Flask App with Pyctuator"\napp = Flask(app_name)\n\n\n@app.route("/")\ndef hello():\n    return "Hello World!"\n\n\nPyctuator(\n    app,\n    app_name,\n    app_url="http://host.docker.internal:5000",\n    pyctuator_endpoint_url="http://host.docker.internal:5000/pyctuator",\n    registration_url="http://localhost:8080/instances"\n)\n\napp.run(debug=False, port=5000)\n```\n\nThe application will automatically register with Spring Boot Admin upon start up.\n\nLog in to the Spring Boot Admin UI at `http://localhost:8080` to interact with the application. \n\n### FastAPI\nThe following example is complete and should run as is.\n\n```python\nfrom fastapi import FastAPI\nfrom uvicorn import Server\n\nfrom uvicorn.config import Config\nfrom pyctuator.pyctuator import Pyctuator\n\n\napp_name = "FastAPI App with Pyctuator"\napp = FastAPI(title=app_name)\n\n\n@app.get("/")\ndef hello():\n    return "Hello World!"\n\n\nPyctuator(\n    app,\n    "FastAPI Pyctuator",\n    app_url="http://host.docker.internal:8000",\n    pyctuator_endpoint_url="http://host.docker.internal:8000/pyctuator",\n    registration_url="http://localhost:8080/instances"\n)\n\nServer(config=(Config(app=app, loop="asyncio"))).run()\n```\n\nThe application will automatically register with Spring Boot Admin upon start up.\n\nLog in to the Spring Boot Admin UI at `http://localhost:8080` to interact with the application. \n\n### aiohttp\nThe following example is complete and should run as is.\n\n```python\nfrom aiohttp import web\nfrom pyctuator.pyctuator import Pyctuator\n\napp = web.Application()\nroutes = web.RouteTableDef()\n\n@routes.get("/")\ndef hello():\n    return web.Response(text="Hello World!")\n\nPyctuator(\n    app,\n    "aiohttp Pyctuator",\n    app_url="http://host.docker.internal:8888",\n    pyctuator_endpoint_url="http://host.docker.internal:8888/pyctuator",\n    registration_url="http://localhost:8080/instances"\n)\n\napp.add_routes(routes)\nweb.run_app(app, port=8888)\n```\n\nThe application will automatically register with Spring Boot Admin upon start up.\n\nLog in to the Spring Boot Admin UI at `http://localhost:8080` to interact with the application.\n\n### Registration Notes\nWhen registering a service in Spring Boot Admin, note that:\n* **Docker** - If the Spring Boot Admin is running in a container while the managed service is running in the docker-host directly, the `app_url` and `pyctuator_endpoint_url` should use `host.docker.internal` as the url\'s host so Spring Boot Admin will be able to connect to the monitored service.\n* **Http Traces** - In order for the "Http Traces" tab to be able to hide requests sent by Spring Boot Admin to the Pyctuator endpoint, `pyctuator_endpoint_url` must be using the same host and port as `app_url`.\n* **HTTPS** - If Pyctuator is to be registered with Spring Boot Admin using HTTPS and the default SSL context is inappropriate, you can provide your own `ssl.SSLContext` using the `ssl_context` optional parameter of the `Pyctuator` constructor.\n* **Insecure HTTPS** - If Spring Boot Admin is using HTTPS with self-signed certificate, set the `PYCTUATOR_REGISTRATION_NO_CERT` environment variable so Pyctuator will disable certificate validation when registering (and deregistering).\n\n## Advanced Configuration\nThe following sections are intended for advanced users who want to configure advanced Pyctuator features.\n\n### Application Info\nWhile Pyctuator only needs to know the application\'s name, we recommend that applications monitored by Spring \nBoot Admin will show additional build and git details. \nThis becomes handy when scaling out a service to multiple instances by showing the version of each instance.\nTo do so, you can provide additional build and git info using methods of the Pyctuator object:\n\n```python\npyctuator = Pyctuator(...)  # arguments removed for brevity\n\npyctuator.set_build_info(\n    name="app",\n    version="1.3.1",\n    time=datetime.fromisoformat("2019-12-21T10:09:54.876091"),\n)\n\npyctuator.set_git_info(\n    commit="7d4fef3",\n    time=datetime.fromisoformat("2019-12-24T14:18:32.123432"),\n    branch="origin/master",\n)\n```\n\nOnce you configure build and git info, you should see them in the Details tab of Spring Boot Admin:\n\n![Detailed Build Info](examples/images/Main_Details_BuildInfo.png)\n\n### Additional Application Info\nIn addition to adding build and git info, Pyctuator allows adding arbitrary application details to the "Info" section in SBA.\n\nThis is done by initializing the `additional_app_info` parameter with an arbitrary dictionary.\nFor example, you can provide links to your application\'s metrics:\n```python\nPyctuator(\n  app,\n  "Flask Pyctuator",\n  app_url=f"http://172.18.0.1:5000",\n  pyctuator_endpoint_url=f"http://172.18.0.1:5000/pyctuator",\n  registration_url=f"http://localhost:8080/instances",\n  app_description="Demonstrate Spring Boot Admin Integration with Flask",\n  additional_app_info=dict(\n    serviceLinks=dict(\n      metrics="http://xyz/service/metrics"\n    ),\n    podLinks=dict(\n      metrics=["http://xyz/pod/metrics/memory", "http://xyz/pod/metrics/cpu"]\n    )\n  )\n)\n```\n\nThis will result with the following Info page in SBA:\n![img.png](examples/images/Additional_App_Info.png)\n\n### DB Health\nFor services that use SQL database via SQLAlchemy, Pyctuator can easily monitor and expose the connection\'s health \nusing the DbHealthProvider class as demonstrated below:\n\n```python\nengine = create_engine("mysql+pymysql://root:root@localhost:3306")\npyctuator = Pyctuator(...)  # arguments removed for brevity\npyctuator.register_health_provider(DbHealthProvider(engine))\n```\n\nOnce you configure the health provider, you should see DB health info in the Details tab of Spring Boot Admin:\n\n![DB Health](examples/images/Main_DB_Health.png)\n\n### Redis health\nIf your service is using Redis, Pyctuator can monitor the connection to Redis by simply initializing a `RedisHealthProvider`:\n\n```python\nr = redis.Redis()\npyctuator = Pyctuator(...)  # arguments removed for brevity\npyctuator.register_health_provider(RedisHealthProvider(r))\n```\n\n### Custom Environment\nOut of the box, Pyctuator exposes Python\'s environment variables to Spring Boot Admin.\n\nIn addition, an application may register an environment provider to provide additional configuration that should be exposed via Spring Boot Admin. \n\nWhen the environment provider is called it should return a dictionary describing the environment. The returned dictionary is exposed to Spring Boot Admin.\n\nSince Spring Boot Admin doesn\'t support hierarchical environment (only a flat key/value mapping), the provided environment is flattened as dot-delimited keys.\n\nPyctuator tries to hide secrets from being exposed to Spring Boot Admin by replacing the values of "suspicious" keys with ***.\n\nSuspicious keys are keys that contain the words "secret", "password" and some forms of "key".\n\nFor example, if an application\'s configuration looks like this:\n\n```python\nconfig = {\n    "a": "s1",\n    "b": {\n        "secret": "ha ha",\n        "c": 625,\n    },\n    "d": {\n        "e": True,\n        "f": "hello",\n        "g": {\n            "h": 123,\n            "i": "abcde"\n        }\n    }\n}\n```\n\nAn environment provider can be registered like so:\n\n```python\npyctuator.register_environment_provider("config", lambda: config)\n```\n\n### Filesystem and Memory Metrics\nPyctuator can provide filesystem and memory metrics.\n\nTo enable these metrics, install [psutil](https://github.com/giampaolo/psutil)\n\nNote that the `psutil` dependency is **optional** and is only required if you want to enable filesystem and memory monitoring.\n\n### Loggers\nPyctuator leverages Python\'s builtin `logging` framework and allows controlling log levels at runtime.\n \nNote that in order to control uvicorn\'s log level, you need to provide a logger object when instantiating it. For example:\n```python\nmyFastAPIServer = Server(\n    config=Config(\n        logger=logging.getLogger("uvi"), \n        app=app, \n        loop="asyncio"\n    )\n)\n```\n\n### Spring Boot Admin Using Basic Authentication\nPyctuator supports registration with Spring Boot Admin that requires basic authentications. The credentials are provided when initializing the Pyctuator instance as follows:\n```python\n# NOTE: Never include secrets in your code !!!\nauth = BasicAuth(os.getenv("sba-username"), os.getenv("sba-password"))\n\nPyctuator(\n    app,\n    "Flask Pyctuator",\n    app_url="http://localhost:5000",\n    pyctuator_endpoint_url=f"http://localhost:5000/pyctuator",\n    registration_url=f"http://spring-boot-admin:8080/instances",\n    registration_auth=auth,\n)\n``` \n\n### Protecting Pyctuator with authentication\nSince there are numerous standard approaches to protect an API, Pyctuator doesn\'t explicitly support any of them. Instead, Pyctuator allows to customize its integration with the web-framework.\nSee the example in [fastapi_with_authentication_example_app.py](examples/FastAPI/fastapi_with_authentication_example_app.py).\n\n## Full blown examples\nThe `examples` folder contains full blown Python projects that are built using [Poetry](https://python-poetry.org/).\n\nTo run these examples, you\'ll need to have Spring Boot Admin running in a local docker container. A Spring Boot Admin Docker image is available [here](https://hub.docker.com/r/michayaak/spring-boot-admin).\n\nUnless the example includes a docker-compose file, you\'ll need to start Spring Boot Admin using docker directly:\n```sh\ndocker run --rm -p 8080:8080 --add-host=host.docker.internal:host-gateway michayaak/spring-boot-admin:2.2.3-1\n```\n(the docker image\'s tag represents the version of Spring Boot Admin, so if you need to use version `2.0.0`, use `michayaak/spring-boot-admin:2.0.0` instead, note it accepts connections on port 8082).\n\nThe examples include\n* [FastAPI Example](examples/FastAPI/README.md) - demonstrates integrating Pyctuator with the FastAPI web framework.\n* [Flask Example](examples/Flask/README.md) - demonstrates integrating Pyctuator with the Flask web framework.\n* [Advanced Example](examples/Advanced/README.md) - demonstrates configuring and using all the advanced features of Pyctuator.\n\n## Contributing\nTo set up a development environment, make sure you have Python 3.7 or newer installed, and run `make bootstrap`.\n\nUse `make check` to run static analysis tools.\n\nUse `make test` to run tests.\n',
    'author': 'Michael Yakobi',
    'author_email': 'michael.yakobi@solaredge.com',
    'maintainer': 'Matan Rubin',
    'maintainer_email': 'matan.rubin@solaredge.com',
    'url': 'https://github.com/SolarEdgeTech/pyctuator',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC
from dataclasses import dataclass
from typing import Any

from pyctuator.impl.pyctuator_impl import PyctuatorImpl


@dataclass
class LinkHref:
    href: str
    templated: bool


# mypy: ignore_errors
# pylint: disable=too-many-instance-attributes
@dataclass
class EndpointsLinks:
    self: LinkHref
    env: LinkHref
    info: LinkHref
    health: LinkHref
    metrics: LinkHref
    loggers: LinkHref
    dump: LinkHref
    threaddump: LinkHref
    logfile: LinkHref
    httptrace: LinkHref


@dataclass
class EndpointsData:
    _links: EndpointsLinks


class PyctuatorRouter(ABC):

    def __init__(
            self,
            app: Any,
            pyctuator_impl: PyctuatorImpl,
    ):
        self.app = app
        self.pyctuator_impl = pyctuator_impl

    def get_endpoints_data(self) -> EndpointsData:
        return EndpointsData(EndpointsLinks(
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url, False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/env", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/info", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/health", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/metrics", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/loggers", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/dump", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/threaddump", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/logfile", False),
            LinkHref(self.pyctuator_impl.pyctuator_endpoint_url + "/httptrace", False),
        ))



from typing import Callable, Dict

from pyctuator.environment.environment_provider import PropertiesSource, EnvironmentProvider, PropertyValue


def _flatten(prefix: str, dict_to_flatten: Dict) -> Dict:
    """
    Recursively flattens a dictionary that may contain literal values (numbers and strings) and other dictionaries.
    For example, given the following dictionary: {
        "a": 1,
        "b": {
            "c": 2
            "d": {
                "e": 3
            }
        }
    }
    The flattened dictionary will be: {
        "a": 1,
        "b.c": 2,
        "b.d.e": 3
    }

    :param prefix: when descending to a sub-dictionary, the prefix represents the keys higher in the hierarchy
    :param dict_to_flatten: a dictionary, or a sub-dictionary to be flattened
    :return: a dictionary from a dot-separated key to a literal value
    """
    res: Dict = {}
    for key, value in dict_to_flatten.items():
        key_with_prefix = f"{prefix}{key}."
        if isinstance(value, dict):
            res = {**res, **_flatten(key_with_prefix, value)}
        else:
            res[key_with_prefix[:-1]] = value
    return res


class CustomEnvironmentProvider(EnvironmentProvider):

    def __init__(self, name: str, env_provider: Callable[[], Dict]) -> None:
        self.name = name
        self.env_provider = env_provider

    def get_properties_source(self, secret_scrubber: Callable[[Dict], Dict]) -> PropertiesSource:
        flattened_env = _flatten("", self.env_provider())
        scrubbed_env = secret_scrubber(flattened_env)
        properties_dict = {key: PropertyValue(value) for (key, value) in scrubbed_env.items()}
        return PropertiesSource(self.name, properties_dict)



import dataclasses
import json
from collections import defaultdict
from datetime import datetime
from functools import partial
from http import HTTPStatus
from typing import Any, Callable, List, Mapping

from aiohttp import web
from multidict import CIMultiDictProxy

from pyctuator.httptrace import TraceRecord, TraceRequest, TraceResponse
from pyctuator.impl import SBA_V2_CONTENT_TYPE
from pyctuator.impl.pyctuator_impl import PyctuatorImpl
from pyctuator.impl.pyctuator_router import PyctuatorRouter


# pylint: disable=too-many-locals,unused-argument
class AioHttpPyctuator(PyctuatorRouter):
    def __init__(self, app: web.Application, pyctuator_impl: PyctuatorImpl) -> None:
        super().__init__(app, pyctuator_impl)

        custom_dumps = partial(
            json.dumps, default=self._custom_json_serializer
        )

        async def empty_handler(request: web.Request) -> web.Response:
            return web.Response(text='')

        async def get_endpoints(request: web.Request) -> web.Response:
            return web.json_response(self.get_endpoints_data(), dumps=custom_dumps)

        async def get_environment(request: web.Request) -> web.Response:
            return web.json_response(pyctuator_impl.get_environment(), dumps=custom_dumps)

        async def get_info(request: web.Request) -> web.Response:
            return web.json_response(pyctuator_impl.get_app_info(), dumps=custom_dumps)

        async def get_health(request: web.Request) -> web.Response:
            return web.json_response(pyctuator_impl.get_health(), dumps=custom_dumps)

        async def get_metric_names(request: web.Request) -> web.Response:
            return web.json_response(pyctuator_impl.get_metric_names(), dumps=custom_dumps)

        async def get_loggers(request: web.Request) -> web.Response:
            return web.json_response(pyctuator_impl.logging.get_loggers(), dumps=custom_dumps)

        async def set_logger_level(request: web.Request) -> web.Response:
            request_dict = await request.json()
            pyctuator_impl.logging.set_logger_level(
                request.match_info["logger_name"],
                request_dict.get("configuredLevel", None),
            )
            return web.json_response({})

        async def get_logger(request: web.Request) -> web.Response:
            logger_name = request.match_info["logger_name"]
            return web.json_response(pyctuator_impl.logging.get_logger(logger_name), dumps=custom_dumps)

        async def get_thread_dump(request: web.Request) -> web.Response:
            return web.json_response(pyctuator_impl.get_thread_dump(), dumps=custom_dumps)

        async def get_httptrace(request: web.Request) -> web.Response:
            raw_data = pyctuator_impl.http_tracer.get_httptrace()
            return web.json_response(raw_data, dumps=custom_dumps)

        async def get_metric_measurement(request: web.Request) -> web.Response:
            return web.json_response(
                pyctuator_impl.get_metric_measurement(request.match_info["metric_name"]),
                dumps=custom_dumps)

        async def get_logfile(request: web.Request) -> web.Response:
            range_header = request.headers.get("range")
            if not range_header:
                return web.Response(
                    body=f"{pyctuator_impl.logfile.log_messages.get_range()}"
                )

            str_res, start, end = pyctuator_impl.logfile.get_logfile(range_header)
            response = web.Response(
                status=HTTPStatus.PARTIAL_CONTENT.value,
                body=str_res,
                headers={
                    "Content-Type": "text/html; charset=UTF-8",
                    "Accept-Ranges": "bytes",
                    "Content-Range": f"bytes {start}-{end}/{end}",
                },
            )
            return response

        @web.middleware
        async def intercept_requests_and_responses(request: web.Request, handler: Callable) -> Any:
            request_time = datetime.now()
            response = await handler(request)
            response_time = datetime.now()

            # Set the SBA-V2 content type for responses from Pyctuator
            if request.url.path.startswith(self.pyctuator_impl.pyctuator_endpoint_path_prefix):
                response.headers["Content-Type"] = SBA_V2_CONTENT_TYPE

            # Record the request and response
            new_record = self._create_record(
                request, response, request_time, response_time
            )
            self.pyctuator_impl.http_tracer.add_record(record=new_record)
            return response

        app.add_routes(
            [
                web.get("/pyctuator", get_endpoints),
                web.options("/pyctuator/env", empty_handler),
                web.options("/pyctuator/info", empty_handler),
                web.options("/pyctuator/health", empty_handler),
                web.options("/pyctuator/metrics", empty_handler),
                web.options("/pyctuator/loggers", empty_handler),
                web.options("/pyctuator/dump", empty_handler),
                web.options("/pyctuator/threaddump", empty_handler),
                web.options("/pyctuator/logfile", empty_handler),
                web.options("/pyctuator/trace", empty_handler),
                web.options("/pyctuator/httptrace", empty_handler),
                web.get("/pyctuator/env", get_environment),
                web.get("/pyctuator/info", get_info),
                web.get("/pyctuator/health", get_health),
                web.get("/pyctuator/metrics", get_metric_names),
                web.get("/pyctuator/metrics/{metric_name}", get_metric_measurement),
                web.get("/pyctuator/loggers", get_loggers),
                web.get("/pyctuator/loggers/{logger_name}", get_logger),
                web.post("/pyctuator/loggers/{logger_name}", set_logger_level),
                web.get("/pyctuator/dump", get_thread_dump),
                web.get("/pyctuator/threaddump", get_thread_dump),
                web.get("/pyctuator/logfile", get_logfile),
                web.get("/pyctuator/trace", get_httptrace),
                web.get("/pyctuator/httptrace", get_httptrace),
            ]
        )
        app.middlewares.append(intercept_requests_and_responses)

    def _custom_json_serializer(self, value: Any) -> Any:
        if dataclasses.is_dataclass(value):
            return dataclasses.asdict(value)

        if isinstance(value, datetime):
            return str(value)
        return None

    def _create_headers_dictionary(self, headers: CIMultiDictProxy[str]) -> Mapping[str, List[str]]:
        headers_dict: Mapping[str, List[str]] = defaultdict(list)
        for (key, value) in headers.items():
            headers_dict[key].append(value)
        return dict(headers_dict)

    def _create_record(
            self,
            request: web.Request,
            response: web.Response,
            request_time: datetime,
            response_time: datetime
    ) -> TraceRecord:
        new_record: TraceRecord = TraceRecord(
            request_time,
            None,
            None,
            TraceRequest(
                request.method,
                str(request.url),
                self._create_headers_dictionary(request.headers),
            ),
            TraceResponse(
                response.status,
                self._create_headers_dictionary(CIMultiDictProxy(response.headers))
            ),
            int((response_time.timestamp() - request_time.timestamp()) * 1000),
        )
        return new_record



import dataclasses
from dataclasses import dataclass
from datetime import datetime
from typing import List, Dict, Mapping, Optional, Callable
from urllib.parse import urlparse

from pyctuator.environment.environment_provider import EnvironmentData, EnvironmentProvider
from pyctuator.environment.scrubber import SecretScrubber
from pyctuator.health.health_provider import HealthStatus, HealthSummary, Status, HealthProvider
from pyctuator.httptrace.http_tracer import HttpTracer
from pyctuator.logfile.logfile import PyctuatorLogfile  # type: ignore
from pyctuator.logging.pyctuator_logging import PyctuatorLogging
from pyctuator.metrics.metrics_provider import Metric, MetricNames, MetricsProvider
from pyctuator.threads.thread_dump_provider import ThreadDump, ThreadDumpProvider


@dataclass
class GitCommitInfo:
    time: datetime
    id: str


@dataclass
class GitInfo:
    commit: GitCommitInfo
    branch: Optional[str] = None


@dataclass
class BuildInfo:
    name: Optional[str] = None
    artifact: Optional[str] = None
    group: Optional[str] = None
    version: Optional[str] = None
    time: Optional[datetime] = None


@dataclass
class AppDetails:
    name: str
    description: Optional[str] = None


@dataclass
class AppInfo:
    app: AppDetails
    build: Optional[BuildInfo] = None
    git: Optional[GitInfo] = None


class PyctuatorImpl:
    # pylint: disable=too-many-instance-attributes
    def __init__(
            self,
            app_info: AppInfo,
            pyctuator_endpoint_url: str,
            logfile_max_size: int,
            logfile_formatter: str,
            additional_app_info: Optional[dict],
    ):
        self.app_info = app_info
        self.pyctuator_endpoint_url = pyctuator_endpoint_url
        self.additional_app_info = additional_app_info

        self.metrics_providers: List[MetricsProvider] = []
        self.health_providers: List[HealthProvider] = []
        self.environment_providers: List[EnvironmentProvider] = []
        self.logging = PyctuatorLogging()
        self.thread_dump_provider = ThreadDumpProvider()
        self.logfile = PyctuatorLogfile(max_size=logfile_max_size, formatter=logfile_formatter)
        self.http_tracer = HttpTracer()

        self.secret_scrubber: Callable[[Dict], Dict] = SecretScrubber().scrub_secrets

        # Determine the endpoint's URL path prefix and make sure it doesn't end with a "/"
        self.pyctuator_endpoint_path_prefix = urlparse(pyctuator_endpoint_url).path
        if self.pyctuator_endpoint_path_prefix[-1:] == "/":
            self.pyctuator_endpoint_path_prefix = self.pyctuator_endpoint_path_prefix[:-1]

    def register_metrics_provider(self, provider: MetricsProvider) -> None:
        self.metrics_providers.append(provider)

    def register_health_providers(self, provider: HealthProvider) -> None:
        self.health_providers.append(provider)

    def register_environment_provider(self, provider: EnvironmentProvider) -> None:
        self.environment_providers.append(provider)

    def set_secret_scrubber(self, secret_scrubber: Callable[[Dict], Dict]) -> None:
        self.secret_scrubber = secret_scrubber

    def get_environment(self) -> EnvironmentData:
        active_profiles: List[str] = list()
        env_data = EnvironmentData(
            active_profiles,
            [source.get_properties_source(self.secret_scrubber) for source in self.environment_providers]
        )
        return env_data

    def set_git_info(self, git_info: GitInfo) -> None:
        self.app_info.git = git_info

    def set_build_info(self, build_info: BuildInfo) -> None:
        self.app_info.build = build_info

    def get_health(self) -> HealthSummary:
        health_statuses: Mapping[str, HealthStatus] = {
            provider.get_name(): provider.get_health()
            for provider in self.health_providers
            if provider.is_supported()
        }

        # Health is UP if no provider is registered
        if not health_statuses:
            return HealthSummary(Status.UP, health_statuses)

        # If there's at least one provider and any of the providers is DOWN, the service is DOWN
        service_is_down = any(health_status.status == Status.DOWN for health_status in health_statuses.values())
        if service_is_down:
            return HealthSummary(Status.DOWN, health_statuses)

        # IF there's at least one provider and none of the providers is DOWN and at least one is UP, the service is UP
        service_is_up = any(health_status.status == Status.UP for health_status in health_statuses.values())
        if service_is_up:
            return HealthSummary(Status.UP, health_statuses)

        # else, all providers are unknown so the service is UNKNOWN
        return HealthSummary(Status.UNKNOWN, health_statuses)

    def get_metric_names(self) -> MetricNames:
        metric_names = []
        for provider in self.metrics_providers:
            for metric_name in provider.get_supported_metric_names():
                metric_names.append(metric_name)
        return MetricNames(metric_names)

    def get_metric_measurement(self, metric_name: str) -> Metric:
        for provider in self.metrics_providers:
            if metric_name.startswith(provider.get_prefix()):
                return provider.get_metric(metric_name)
        raise KeyError(f"Unknown metric {metric_name}")

    def get_thread_dump(self) -> ThreadDump:
        return self.thread_dump_provider.get_thread_dump()

    def get_app_info(self) -> Dict:
        app_info_dict = {k: v for (k, v) in dataclasses.asdict(self.app_info).items() if v}

        if self.additional_app_info:
            app_info_dict = {**app_info_dict, **self.additional_app_info}

        return app_info_dict
